from typing import Any, Dict

import oaas._registrations as registrations
import oaas_simple.server.server
from oaas import service

from oaas_simple.data import from_data, create_data
from oaas_simple.rpc import call_pb2
from oaas_simple.rpc import call_pb2_grpc


@service("service-invoker-proxy")
class ServiceInvokerProxy(call_pb2_grpc.ServiceInvokerServicer):
    def __init__(self) -> None:
        self._service_instances: Dict[str, Any] = dict()

        server = find_simple_server()

        # we only register the non grpc services as simple services
        for service_definition in registrations.services:
            if server._grpc_server.can_serve(service_definition):
                continue

            self._service_instances[service_definition.name] = service_definition.code()

    """
    Invokes the service on this server.
    """

    def InvokeMethod(self, request: call_pb2.ServiceCall, context) -> call_pb2.Data:  # type: ignore
        service_instance = self._service_instances[request.service]  # type: ignore

        service_method = getattr(service_instance, request.method)  # type: ignore

        args = []
        kw = {}

        for param in request.parameters:  # type: ignore
            value = from_data(param.data)

            if param.name:
                kw[param.name] = value
            else:
                args.append(value)

        result = service_method(*args, **kw)

        return create_data(result)


def find_simple_server():
    for server in registrations.servers_middleware:
        if isinstance(server, oaas_simple.server.server.OaasSimpleServer):
            return server

    raise Exception("No OaasSimpleServer was registered")


def noop():
    pass



import oaas
from oaas_grpc.server import OaasGrpcServer
from oaas_grpc.server.find_ips import find_ips
from oaas_registry_api import OaasRegistryStub
from oaas_registry_api.rpc.registry_pb2 import OaasServiceDefinition
from oaas_simple.server.service_invoker_proxy import noop

noop()


class OaasSimpleServer(oaas.ServerMiddleware):
    def __init__(self, *, port=8999):
        super(OaasSimpleServer, self).__init__()

        self._grpc_server = OaasGrpcServer(port=port)
        self.port = port

    def serve(self) -> None:
        # we ensure we're serving with the gRPC server first
        self._grpc_server.serve()

        # we can get the registry only after the server is servicing
        # in case this _is_ the registry.
        registry = oaas.get_client(OaasRegistryStub)

        # we register the services
        locations = find_ips(port=self.port)
        for service_definition in oaas.registrations.services:
            if self._grpc_server.can_serve(service_definition=service_definition):
                continue

            print(
                f"Added SIMPLE service: {service_definition.gav} as {service_definition.code}"
            )

            registry.register_service(
                OaasServiceDefinition(
                    namespace=service_definition.namespace,
                    name=service_definition.name,
                    version=service_definition.version,
                    tags={
                        "_protocol": "simple",
                    },
                    locations=locations,
                )
            )

    def join(self) -> None:
        self._grpc_server.join()

    def can_serve(self, service_definition: oaas.ServiceDefinition) -> bool:
        # We can serve both gRPC, and simple services, since we embed both
        return True



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import oaas_simple.rpc.call_pb2 as call__pb2


class ServiceInvokerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.InvokeMethod = channel.unary_unary(
            "/gr.ServiceInvoker/InvokeMethod",
            request_serializer=call__pb2.ServiceCall.SerializeToString,
            response_deserializer=call__pb2.Data.FromString,
        )


class ServiceInvokerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def InvokeMethod(self, request, context):
        """*
        Invoke a single method.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    @staticmethod
    def add_to_server(servicer, server):
        rpc_method_handlers = {
            "InvokeMethod": grpc.unary_unary_rpc_method_handler(
                servicer.InvokeMethod,
                request_deserializer=call__pb2.ServiceCall.FromString,
                response_serializer=call__pb2.Data.SerializeToString,
            ),
        }
        generic_handler = grpc.method_handlers_generic_handler(
            "gr.ServiceInvoker", rpc_method_handlers
        )
        server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class ServiceInvoker(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def InvokeMethod(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/gr.ServiceInvoker/InvokeMethod",
            call__pb2.ServiceCall.SerializeToString,
            call__pb2.Data.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from typing import Dict, Optional

import oaas
import oaas.registry


@oaas.service("oaas-registry")
class OaasRegistryService(oaas.registry.OaasRegistry):
    # FIXME: externalize into its own package? or move to core OaaS?
    def __init__(self) -> None:
        self._services: Dict[str, oaas.registry.ServiceAddress] = dict()

    def resolve_service(
        self, service_definition: oaas.registry.ServiceDefinition
    ) -> Optional[oaas.registry.ServiceAddress]:
        return self._services.get(service_definition_key(service_definition), None)

    def register_service(
        self,
        service_definition: oaas.registry.ServiceDefinition,
        service_address: oaas.registry.ServiceAddress,
    ) -> None:
        self._services[service_definition_key(service_definition)] = service_address


def service_definition_key(service_definition: oaas.registry.ServiceDefinition) -> str:
    return (
        service_definition.get("protocol", "")
        + ":"
        + service_definition.get("namespace", "")
        + ":"
        + service_definition.get("name", "")
        + ":"
        + service_definition.get("version", "")
    )



import socket
from typing import Any, Optional

import oaas
from oaas import ClientDefinition
from oaas_grpc.client.client import OaasGrpcClient
from oaas_simple.client.service_client_proxy import ServiceClientProxy
from oaas_simple.rpc.call_pb2_grpc import ServiceInvokerStub


def is_someone_listening(host_address: str, port: int):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as a_socket:
        location = (host_address, port)
        try:
            result_of_check = a_socket.connect_ex(location)

            return result_of_check == 0
        except Exception:
            return False


class OaasSimpleClient(oaas.ClientMiddleware):
    def __init__(self) -> None:
        self._grpc_client: Optional[OaasGrpcClient] = None

        # FIXME: instantiating a client shouldn't magically add it
        oaas.register_client_provider(OaasGrpcClient())

    def create_client(self, client_definition: ClientDefinition) -> Any:
        client_proxy = self._oaas_grpc_client.create_grpc_client(
            namespace=client_definition.namespace,
            name=client_definition.name,
            version=client_definition.version,
            tags=client_definition.tags,
            code=ServiceInvokerStub,
        )

        return ServiceClientProxy(
            client_definition=client_definition, stub=client_proxy
        )

    def can_handle(self, client_definition: ClientDefinition) -> bool:
        """
        We only accept services that end with Stub, that are
        probably generated by gRPC.
        """
        name = client_definition.code.__name__

        return not name.endswith("Stub")

    @property
    def _oaas_grpc_client(self) -> OaasGrpcClient:
        if self._grpc_client:
            return self._grpc_client

        for client in oaas.registrations.clients_middleware:
            if isinstance(client, OaasGrpcClient):
                self._grpc_client = client
                return self._grpc_client

        self._grpc_client = OaasGrpcClient()
        return self._grpc_client



import json
from typing import Union, Any

from oaas_simple.rpc import call_pb2


def create_data(data: Union[str, bytes, Any]) -> call_pb2.Data:
    if isinstance(data, str):
        return call_pb2.Data(s=data)

    if isinstance(data, bytes):
        return call_pb2.Data(b=data)

    json_data = json.dumps(data)
    return call_pb2.Data(json=json_data)


def from_data(data: call_pb2.Data) -> Union[Any, str, bytes]:
    if data.s:
        return data.s

    if data.b:
        return data.b

    return json.loads(data.json)



from typing import Any, Dict, Callable

from oaas import ClientDefinition

from oaas_simple.client.single_method import SingleMethod
from oaas_simple.rpc import call_pb2_grpc


class ServiceClientProxy:
    """
    Makes it so any call invoked on it, is being forwarded
    to the service.
    """

    def __init__(
        self,
        *,
        client_definition: ClientDefinition,
        stub: call_pb2_grpc.ServiceInvokerStub,
    ) -> None:
        self._methods: Dict[str, Callable] = dict()
        self.client_definition = client_definition
        self.stub = stub

    def __getattr__(self, method_name: str):
        result = self._methods.get(method_name, None)

        if result:
            return result

        invoker = SingleMethod(
            stub=self.stub,
            client_definition=self.client_definition,
            method_name=method_name,
        )

        self._methods[method_name] = invoker

        return invoker



import oaas

from oaas_simple.registry.local_registry import OaasRegistryService
from oaas_simple.server.server import OaasSimpleServer

# noop on purpose, just to register it
dir(OaasRegistryService)

server_instance = OaasSimpleServer()
oaas.register_server_provider(server_instance)

oaas.serve()
print(f"OaaS Simple Registry started on port {server_instance.port}.")
oaas.join()
